"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parse = exports.normalize = exports.defaultComposition = void 0;
const _ = require("lodash");
const path = require("path");
const errors_1 = require("./errors");
const schemas_1 = require("./schemas");
const packageJson = require('../package.json');
const packageVersion = packageJson.version;
function defaultComposition(image, dockerfile) {
    let context;
    if (image) {
        context = `image: ${image}`;
    }
    else {
        if (dockerfile) {
            context = `build: {context: ".", dockerfile: "${dockerfile}"}`;
        }
        else {
            context = 'build: "."';
        }
    }
    return `# Auto-generated compose file by resin-compose-parse@v${packageVersion}
version: '${schemas_1.DEFAULT_SCHEMA_VERSION}'
networks: {}
volumes:
  resin-data: {}
services:
  main:
    ${context}
    privileged: true
    tty: true
    restart: always
    network_mode: host
    volumes:
      - type: volume
        source: resin-data
        target: /data
    labels:
      io.resin.features.kernel-modules: 1
      io.resin.features.firmware: 1
      io.resin.features.dbus: 1
      io.resin.features.supervisor-api: 1
      io.resin.features.resin-api: 1
`;
}
exports.defaultComposition = defaultComposition;
function normalize(inputCompositionObject, fileResolverCb) {
    if (fileResolverCb === undefined) {
        return normalizeObjectToComposition(inputCompositionObject);
    }
    else {
        const composition = normalizeObjectToComposition(inputCompositionObject);
        return expandContext(composition, fileResolverCb);
    }
}
exports.normalize = normalize;
function normalizeObjectToComposition(inputCompositionObject) {
    if (!_.isObject(inputCompositionObject)) {
        throw new errors_1.ValidationError('Invalid composition format');
    }
    let version;
    let c = inputCompositionObject;
    if (_.isUndefined(c.version)) {
        version = schemas_1.SchemaVersion.v1;
    }
    else {
        if (!_.isString(c.version)) {
            c.version = `${c.version}`;
        }
        switch (c.version) {
            case '2':
            case '2.0':
            case '2.1':
            case '2.2':
            case '2.3':
            case '2.4':
                version = schemas_1.DEFAULT_SCHEMA_VERSION;
                break;
            default:
                throw new errors_1.ValidationError('Unsupported composition version');
        }
    }
    preflight(version, c);
    try {
        schemas_1.validate(version, c);
    }
    catch (e) {
        if (e instanceof schemas_1.SchemaError) {
            throw new errors_1.ValidationError(e);
        }
        throw e;
    }
    switch (version) {
        case schemas_1.SchemaVersion.v1:
            c = { version: schemas_1.DEFAULT_SCHEMA_VERSION, services: c };
        case schemas_1.DEFAULT_SCHEMA_VERSION:
            if (c.volumes) {
                c.volumes = _.mapValues(c.volumes, normalizeVolume);
            }
            if (c.networks) {
                c.networks = _.mapValues(c.networks, normalizeNetwork);
            }
            const services = c.services || {};
            const serviceNames = _.keys(services);
            const volumeNames = _.keys(c.volumes);
            const networkNames = _.keys(c.networks);
            c.services = _(services)
                .map((service, serviceName) => {
                try {
                    const normalizedService = normalizeService(service, serviceNames, volumeNames, networkNames);
                    return [serviceName, normalizedService];
                }
                catch (err) {
                    if (err instanceof errors_1.ValidationError) {
                        throw new errors_1.ServiceError(serviceName, err);
                    }
                    throw err;
                }
            })
                .fromPairs()
                .value();
    }
    c.version = schemas_1.DEFAULT_SCHEMA_VERSION;
    return c;
}
function preflight(_version, data) {
    if (_.isObject(data.networks)) {
        data.networks = _.mapValues(data.networks, (n) => n || {});
    }
    if (_.isObject(data.volumes)) {
        data.volumes = _.mapValues(data.volumes, (v) => v || {});
    }
}
function normalizeService(service, serviceNames, volumeNames, networkNames) {
    if (!service.image && !service.build) {
        throw new errors_1.ValidationError('You must specify either an image or a build');
    }
    if (service.build) {
        service.build = normalizeServiceBuild(service.build, networkNames);
    }
    if (service.depends_on) {
        if (!_.isArray(service.depends_on)) {
            service.depends_on = _.map(service.depends_on, (dep, serviceName) => {
                if (_.includes(['service_started', 'service-started'], dep.condition)) {
                    return serviceName;
                }
                throw new errors_1.ValidationError('Only "service_started" type of service dependency is supported');
            });
        }
        if (_.uniq(service.depends_on).length !== service.depends_on.length) {
            throw new errors_1.ValidationError('Service dependencies must be unique');
        }
        _.forEach(service.depends_on, (dep) => {
            if (!_.includes(serviceNames, dep)) {
                throw new errors_1.ValidationError(`Unknown service dependency: ${dep}`);
            }
        });
    }
    if (service.environment) {
        service.environment = normalizeKeyValuePairs(service.environment);
    }
    if (service.env_file) {
        service.env_file = normalizeAndValidateFilePath(service.env_file);
    }
    if (service.extra_hosts) {
        if (!_.isArray(service.extra_hosts)) {
            service.extra_hosts = normalizeExtraHostObject(service.extra_hosts);
        }
    }
    if (service.labels) {
        service.labels = normalizeKeyValuePairs(service.labels);
        validateLabels(service.labels);
    }
    if (service.ports) {
        service.ports = normalizeArrayOfStrings(service.ports);
    }
    if (service.volumes) {
        const [volumes, tmpfs] = normalizeServiceVolumes(service.volumes, volumeNames);
        service.volumes = volumes;
        if (service.tmpfs) {
            if (typeof service.tmpfs === 'string') {
                service.tmpfs = [service.tmpfs].concat(tmpfs);
            }
            else {
                service.tmpfs = service.tmpfs.concat(tmpfs);
            }
        }
    }
    if (service.scale) {
        throw new errors_1.ValidationError('service.scale is not allowed');
    }
    return service;
}
function normalizeArrayOfStrings(value) {
    return _.map(value, String);
}
function normalizeServiceBuild(serviceBuild, networkNames) {
    if (typeof serviceBuild === 'string') {
        serviceBuild = { context: serviceBuild };
    }
    if (serviceBuild.args) {
        serviceBuild.args = normalizeKeyValuePairs(serviceBuild.args);
    }
    if (serviceBuild.labels) {
        serviceBuild.labels = normalizeKeyValuePairs(serviceBuild.labels);
        validateLabels(serviceBuild.labels);
    }
    if (serviceBuild.extra_hosts && !_.isArray(serviceBuild.extra_hosts)) {
        serviceBuild.extra_hosts = normalizeExtraHostObject(serviceBuild.extra_hosts);
    }
    if (serviceBuild.isolation) {
        throw new errors_1.ValidationError('service.build.isolation is not allowed');
    }
    if (serviceBuild.network &&
        serviceBuild.network !== 'host' &&
        serviceBuild.network !== 'none') {
        if (networkNames.indexOf(serviceBuild.network) === -1) {
            throw new errors_1.ValidationError(`Missing network definition for '${serviceBuild.network}'`);
        }
    }
    return serviceBuild;
}
function normalizeServiceVolumes(serviceVolumes, volumeNames) {
    const tmpfs = [];
    const volumes = [];
    serviceVolumes.map((volume) => {
        const ref = normalizeServiceVolume(volume);
        validateServiceVolume(ref, volumeNames);
        switch (ref.type) {
            case 'tmpfs':
                if (ref.target) {
                    tmpfs.push(ref.target);
                }
                break;
            case 'volume':
                volumes.push(`${ref.source}:${ref.target}${ref.read_only ? ':ro' : ''}`);
                break;
        }
    });
    return [volumes, tmpfs];
}
function normalizeServiceVolume(serviceVolume) {
    let ref = { type: 'volume', read_only: false };
    if (typeof serviceVolume === 'string') {
        const parts = serviceVolume.split(':');
        if (parts.length < 2) {
            throw new errors_1.ValidationError(`Invalid volume: '${serviceVolume}'`);
        }
        ref.source = parts[0];
        ref.target = parts[1];
        if (path.parse(ref.source).dir !== '') {
            ref.type = 'bind';
        }
        if (parts[2] === 'ro') {
            ref.read_only = true;
        }
    }
    else {
        ref = serviceVolume;
    }
    return ref;
}
function validateServiceVolume(serviceVolume, volumeNames) {
    switch (serviceVolume.type) {
        case 'bind':
            throw new errors_1.ValidationError('Bind mounts are not allowed');
        case 'tmpfs':
            if (serviceVolume.source) {
                throw new errors_1.ValidationError('Tmpfs mount can not have a source');
            }
            if (!serviceVolume.target) {
                throw new errors_1.ValidationError('Tmpfs mount missing target');
            }
            if (serviceVolume.read_only) {
                throw new errors_1.ValidationError('Tmpfs can not be read only');
            }
            if (serviceVolume.tmpfs) {
                throw new errors_1.ValidationError('Tmpfs options are not allowed');
            }
            break;
        case 'volume':
            if (!serviceVolume.source) {
                throw new errors_1.ValidationError('Missing volume source');
            }
            if (volumeNames.indexOf(serviceVolume.source) === -1) {
                throw new errors_1.ValidationError(`Missing volume definition for '${serviceVolume.source}'`);
            }
            if (serviceVolume.volume) {
                throw new errors_1.ValidationError('Volume options are not allowed');
            }
            break;
    }
}
function validateLabels(labels) {
    _.keys(labels).forEach((name) => {
        if (!/^[a-zA-Z0-9.-]+$/.test(name)) {
            throw new errors_1.ValidationError(`Invalid label name: "${name}". ` +
                'Label names must only contain alphanumeric ' +
                'characters, periods "." and dashes "-".');
        }
    });
}
function normalizeNetwork(network) {
    if (network.labels) {
        network.labels = normalizeKeyValuePairs(network.labels);
        validateLabels(network.labels);
    }
    return network;
}
function normalizeVolume(volume) {
    if (volume.labels) {
        volume.labels = normalizeKeyValuePairs(volume.labels);
        validateLabels(volume.labels);
    }
    return volume;
}
function normalizeExtraHostObject(extraHostsObject) {
    return _.map(extraHostsObject, (ip, host) => `${host}:${ip}`);
}
function parse(c) {
    if (c.version !== schemas_1.DEFAULT_SCHEMA_VERSION) {
        throw new Error('Unsupported composition version');
    }
    return _.toPairs(c.services).map(([name, service]) => {
        return createImageDescriptor(name, service);
    });
}
exports.parse = parse;
function createImageDescriptor(serviceName, service) {
    if (service.image && !service.build) {
        return { serviceName, image: service.image };
    }
    if (!service.build) {
        throw new errors_1.InternalInconsistencyError();
    }
    const build = service.build;
    if (service.image) {
        build.tag = service.image;
    }
    return { serviceName, image: build };
}
function normalizeKeyValuePairs(obj, sep = '=') {
    if (!obj) {
        return {};
    }
    if (!_.isArray(obj)) {
        return _(obj)
            .toPairs()
            .map(([key, value]) => {
            return [key, value ? ('' + value).trim() : ''];
        })
            .fromPairs()
            .value();
    }
    return _(obj)
        .map((val) => {
        const parts = val.split(sep);
        return [parts.shift(), parts.join('=')];
    })
        .map(([key, value]) => {
        return [key.trim(), value ? value.trim() : ''];
    })
        .fromPairs()
        .value();
}
function normalizeAndValidateFilePath(envFile) {
    const normalizedEnvFilePaths = new Set();
    if (!Array.isArray(envFile)) {
        envFile = [envFile];
    }
    for (let envFilePath of envFile) {
        envFilePath = path.normalize(envFilePath);
        if (path.isAbsolute(envFilePath)) {
            throw new errors_1.ValidationError(`Absolute filepath not allowed: ${envFilePath}`);
        }
        if (envFilePath.startsWith('..')) {
            throw new errors_1.ValidationError(`Directory traversing not allowed : ${envFilePath}`);
        }
        if (envFilePath.includes('*')) {
            throw new errors_1.ValidationError(`Wildcards not allowed : ${envFilePath}`);
        }
        normalizedEnvFilePaths.add(envFilePath);
    }
    return [...normalizedEnvFilePaths];
}
function expandContext(composition, fileResolverCb) {
    return __awaiter(this, void 0, void 0, function* () {
        const expandedEnvironmentFiles = yield readEnvFilesFromComposition(composition, fileResolverCb);
        assignExpandedEnvFilesToComposition(composition, expandedEnvironmentFiles);
        return composition;
    });
}
function readEnvFilesFromComposition(composition, fileResolverCb) {
    return __awaiter(this, void 0, void 0, function* () {
        const envFileVariables = {};
        for (const service of Object.values(composition.services)) {
            let envFilePaths = service.env_file;
            if (!!envFilePaths) {
                if (!Array.isArray(envFilePaths)) {
                    envFilePaths = [envFilePaths];
                }
                for (const envFilePath of envFilePaths) {
                    if (!(envFilePath in envFileVariables)) {
                        envFileVariables[envFilePath] = yield readAndNormalizeExpandEnvFile(envFilePath, fileResolverCb);
                    }
                }
            }
        }
        return envFileVariables;
    });
}
function assignExpandedEnvFilesToComposition(composition, envFileVariables) {
    var _a;
    for (const service of Object.values(composition.services)) {
        let envFilePaths = service.env_file;
        if (!!envFilePaths) {
            service.environment = (_a = service.environment) !== null && _a !== void 0 ? _a : {};
            if (!Array.isArray(envFilePaths)) {
                envFilePaths = [envFilePaths];
            }
            for (const envFilePath of envFilePaths) {
                for (const [key, value] of Object.entries(envFileVariables[envFilePath])) {
                    if (!(key in service.environment)) {
                        service.environment[key] = value;
                    }
                }
            }
        }
        delete service.env_file;
    }
    return composition;
}
function readAndNormalizeExpandEnvFile(envFile, fileResolverCb) {
    return __awaiter(this, void 0, void 0, function* () {
        const readline = require('readline');
        const { once } = require('events');
        const intermediateEnv = {};
        let readableError;
        const readable = yield fileResolverCb(envFile);
        const lineReader = readline.createInterface({
            input: readable,
            crlfDelay: Infinity,
        });
        readable.on('error', (readError) => {
            readableError = readError;
            lineReader.close();
        });
        lineReader.on('line', (line) => {
            for (const [key, val] of Object.entries(normalizeKeyValuePairs([line]))) {
                intermediateEnv[key] = val;
            }
        });
        yield once(lineReader, 'close');
        if (readableError !== undefined) {
            throw readableError;
        }
        return intermediateEnv;
    });
}
//# sourceMappingURL=compose.js.map